"""Second Stage Morphologically Aware Learner for simple Klamath data including the suffixes -Ca and -tk.
This includes potential URs mapped out by handcount of forms where the given UR could serve.
The weights are given from running Ca-tk-FirstStage
Charlie O'Hara 2/5/17
"""

import numpy as np
from math import exp, log
import scipy.optimize
import json
import random

#Import data from text file.
with open ("just-tk-Ca.txt", "r") as myfile:
    data=myfile.read().splitlines()

#Dictionary of Morpheme tags and the possible URs associated with each one. 
#With Abstract URs
LEX={'58': {'wen': 2}, '51': {'5wan5a': 1, '5wan5': 2}, '43': {'3u:5a': 1, '3u:5': 2}, '60': {'xiwx': 2, 'xiwxa': 1}, '61': {'s7ew*i': 2}, '62': {'sla:6i': 2}, '63': {')allwi': 2}, '64': {'sdup(i': 2}, '65': {'wejli': 2}, '66': {'gi': 2}, '52': {'da5': 2, 'da5a': 1}, '24': {'s3u5': 2, 's3u5a': 1}, '25': {'?e:)i': 1, '?e:)': 2, '?e:)e': 1, '?e:)I': 1}, '21': {'ncu5a': 1, 'ncu5': 2}, '48': {'*uli': 1, '*ulI': 1, '*ule': 1, '*ul': 2}, '23': {'jle:4': 2, 'jle:4a': 1}, '46': {'3a0ig': 2, '3a0iga': 1}, '47': {'s4a:)a': 1, 's4a:)': 2}, '44': {'su6al0': 2, 'su6al0a': 1}, '45': {'s8e5a': 1, 's8e5': 2}, '42': {'n2u1': 2, 'n2u1a': 1}, '29': {'sqalb': 2, 'sqalba': 1}, '40': {'q1a8': 2, 'q1a8a': 1}, '41': {'wgi1': 2, 'wgi1a': 1}, '1': {'Ca': 49}, '0': {'etk': 5, 'Itk': 5, 'itk': 5, 'tk': 50, 'atk': 38}, '2': {'tsa:4': 2, 'tsa:4a': 1}, '5': {'lu:*': 2, 'lu:*a': 1}, '4': {'6a:si': 1, '6a:sI': 1, '6a:se': 1, '6a:s': 2}, '7': {'4uyu:mI': 1, '4uyu:m': 2, '4uyu:mi': 1, '4uyu:me': 1}, '8': {'m1uqamna': 1, 'm1uqamn': 2}, '38': {'w8i:5': 2, 'w8i:5a': 1}, '59': {'3uc3': 2, '3uc3a': 1}, '14': {'stusq': 2, 'stusqa': 1}, '11': {'mu4': 2, 'mu4a': 1}, '10': {'gmuc': 2, 'gmuca': 1}, '39': {'5u3a': 1, '5u3': 2}, '12': {'nqu2': 2, 'nqu2a': 1}, '15': {'gu)': 2, 'gu)a': 1}, '22': {'l?ega': 1, 'l?eg': 2}, '16': {'gu9': 2, 'gu9a': 1}, '19': {'s?uys?a': 1, 's?uys?': 2}, '54': {'?ewwa': 2, '?eww': 2}, '57': {'pipi:4a': 1, 'pipi:4': 2}, '30': {'xleg': 2, 'xlega': 1}, '37': {'lmu(a': 1, 'lmu(': 2}, '36': {'6e5': 2, '6e5a': 1}, '35': {'se:s': 2, 'se:sa': 1}, '34': {'5e:8': 2, '5e:8i': 1, '5e:8e': 1, '5e:8I': 1}, '55': {'s0enga': 1, 's0eng': 2}, '18': {'s0elw': 2, 's0elwa': 1}}
#Only Concrete URs
LEX={'58': {'wen': 2}, '51': {'5wan5a': 1, '5wan5': 2}, '43': {'3u:5a': 1, '3u:5': 2}, '60': {'xiwx': 2, 'xiwxa': 1}, '61': {'s7ew*i': 2}, '62': {'sla:6i': 2}, '63': {')allwi': 2}, '64': {'sdup(i': 2}, '65': {'wejli': 2}, '66': {'gi': 2}, '52': {'da5': 2, 'da5a': 1}, '24': {'s3u5': 2, 's3u5a': 1}, '25': {'?e:)i': 1, '?e:)': 2}, '21': {'ncu5a': 1, 'ncu5': 2}, '48': {'*uli': 1, '*ul': 2}, '23': {'jle:4': 2, 'jle:4a': 1}, '46': {'3a0ig': 2, '3a0iga': 1}, '47': {'s4a:)a': 1, 's4a:)': 2}, '44': {'su6al0': 2, 'su6al0a': 1}, '45': {'s8e5a': 1, 's8e5': 2}, '42': {'n2u1': 2, 'n2u1a': 1}, '29': {'sqalb': 2, 'sqalba': 1}, '40': {'q1a8': 2, 'q1a8a': 1}, '41': {'wgi1': 2, 'wgi1a': 1}, '1': {'Ca': 49}, '0': {'itk': 5, 'tk': 50, 'atk': 38}, '2': {'tsa:4': 2, 'tsa:4a': 1}, '5': {'lu:*': 2, 'lu:*a': 1}, '4': {'6a:si': 1, '6a:s': 2}, '7': {'4uyu:m': 2, '4uyu:mi': 1}, '8': {'m1uqamna': 1, 'm1uqamn': 2}, '38': {'w8i:5': 2, 'w8i:5a': 1}, '59': {'3uc3': 2, '3uc3a': 1}, '14': {'stusq': 2, 'stusqa': 1}, '11': {'mu4': 2, 'mu4a': 1}, '10': {'gmuc': 2, 'gmuca': 1}, '39': {'5u3a': 1, '5u3': 2}, '12': {'nqu2': 2, 'nqu2a': 1}, '15': {'gu)': 2, 'gu)a': 1}, '22': {'l?ega': 1, 'l?eg': 2}, '16': {'gu9': 2, 'gu9a': 1}, '19': {'s?uys?a': 1, 's?uys?': 2}, '54': {'?ewwa': 2, '?eww': 2}, '57': {'pipi:4a': 1, 'pipi:4': 2}, '30': {'xleg': 2, 'xlega': 1}, '37': {'lmu(a': 1, 'lmu(': 2}, '36': {'6e5': 2, '6e5a': 1}, '35': {'se:s': 2, 'se:sa': 1}, '34': {'5e:8': 2, '5e:8i': 1}, '55': {'s0enga': 1, 's0eng': 2}, '18': {'s0elw': 2, 's0elwa': 1}}


#Set of vowels considered with +high
HIGH=["i","I"]
#Set of vowels with -high
LOW=["a","e"]
#Set of vowels overall
VOWELS="a e i u I".split()
#Set of vowels, including long vowels
VOWLONGS="a e i u a: e: i: u:".split()
#List of the names of the constraints used in this simulation
CONSTRAINTS="Idhi maxv *e maxvin IdATR depv *I *Ctk IdHiin IdATRin".split()
#CONSTRAINTS="Idhi maxv *e maxvin IdATR depv *I *Ctk IdHiin IdATRin".split()
#Initial weights of constraints (here after phonotactic learning)

WEIGHTS=[45.69128685, 45.31034553, 85.43973882, 45.31034551, 0, 8.16611561, 100, 100, 45.69128685, 0]
    #[40.128066262168879, 40.105031430172986, 74.431233044725815, 40.105031430172723, 40.0, 8.2561096115072168, 79.571548958254908, 100.00003197782466, 100.00003197782466, 40.128066262168481, 0.0]
#Whether the constraints are faithfulness (f) or markedness (m), ties to whether they are biased high or low.  
CONTYPE="f f m f f f m m f f".split()

#numcon represents the number of constraints in the grammar.
numcon=len(CONTYPE)
#SIGMAM is a constant that represents the plasticity of markedness constraints, the higher SIGMAM, the more markedness constraints can move.
SIGMAM=10
#SIGMAF does the same thing for faithfulness constraints.
SIGMAF=25
#mbias is the constraint weight that markedness constraints are biased towards---meant to be high, in order to cause restrictiveness
mbias=100
#sigma is used as a scaling factor for the prior which prevents frequencies of URs from getting too high.
sigma=.001
#Bound used to prevent constraint weights and UR frequencies from getting below too negative.
negative=-0.00000001


#Quick allows simulations to be run with just one form of each morpheme type; speeding up simulations for debugging.
quick=True
quick=False


#Set the probability of each underlying form initially equal. 
for x in LEX:
	total=0
	for y in LEX[x]:
		LEX[x][y]=1
		total+=LEX[x][y]
	for y in LEX[x]:
		LEX[x][y]=float(LEX[x][y])/total
		
#Given a Lexicon with frequency counts for each underlying form, returns the probability of each underlying form.		
def pLEX(thing):
	
	tLEX={}
	for x in thing:
		tLEX[x]={}
		for y in thing[x]:

			tLEX[x][y]=0
			tLEX[x][y]=thing[x][y]
	for x in tLEX:
		total=0
		for y in tLEX[x]:
			total+=abs(float(tLEX[x][y]))

		for y in tLEX[x]:
			
				
			tLEX[x][y]=float(tLEX[x][y])/total
	return tLEX

#print LEX

#Make a small dictionary for tests. If quick the learner uses only one form of each type (a-epenthesis, nonalternating i, n-final, and abstract)
if quick:
	small={}
	for x in LEX:
		if x in "58 51 61 25 0 1 3".split():
		#if x in range(30):
			#print LEX
			small[x]=LEX[x]

	LEX=small
	print LEX
	
# Create lexvect and lexvectkey
# lexvect is a list of probabilities
# lexvectkey is a list of tag-UR pairs corresponding to lexvect
lexvect=[]
lexvectkey=[]
for x in LEX:
	for y in LEX[x]:
		lexvect.append(LEX[x][y])
		lexvectkey.append([x,y])

smallvect=[]
smallvectkey=[]
			

#Fill inpout (input-output) with forms from data
#inpout is a dictionary that associates ordered pairs of morpheme tags with their surface forms. 
# (Stem-tag, Suffix-tag) : Surface-Form
inpout={}
for x in data:
	
	pair=x.split('\t')[1].replace('(','').replace(')','').split(', ')
	if quick:
		if pair[0] in "58 51 61 25 0 1".split():
		#if pair[0] in range(30):
			inpout[(pair[0],pair[1])]=x.split('\t')[0]
	else:
		inpout[(pair[0].split()[0],pair[1].split()[0])]=x.split('\t')[0]





#Take in input (ordered pair of stem and suffix URs) , output (ordered pair of surface form and Correspondence relation) and WEIGHTS to find harmony score.
def harmony(input, outputpair, WEIGHTS):
	harmonyscore=0
	#find the violation vector of this input-output candidate
	viols=vios(input[0],input[1], outputpair)

	#Perform the dot product of the constraint weight and violation vectors.
	for x in range(len(viols)):
		harmonyscore+=WEIGHTS[x]*(-viols[x])
	#return harmony score
	return harmonyscore

# Creates a list of parameters for the optimize function, first the weights of the constraints, then the frequencies of the URs. 
def wToParam(weights,lexvect,lexvectkey):
	parm=[]
	for w in weights:
		parm.append(w)
	for i in lexvect:
		parm.append(i)
	return [parm,lexvectkey]

#Create candidates that raise /e/ and /I/ to [i]. Take in an input string, FFC (a fully faithful correspondence relation), and a list of the indices of /e/ and /I/ in the input
def rise(input,FFC,lows):
	#Find the first index of a low vowel. 
	first=lows[0]
	#If there is just one index in lows, return the fully faithful candidate as well as the candidate that replaces the vowel at first with [i]
	if len(lows)==1:
		high="i"
		return [(input,FFC), ("%s%s%s"%(input[:first],high,input[first+1:]),FFC)]
	#If there are multiple indices in lows, this function runs recursively.
	else:
		high="i"
		#Return the union of the candidates that do not raise the vowel at the index first, and those that do; but remove the index from first.
		return rise(input,FFC,lows[1:])+ rise("%s%s%s"%(input[:first],high,input[first+1:]),FFC,lows[1:])
#Create candidates that lower /i/ and /I/ to [e]. Take in an input string, FFC (a fully faithful correspondence relation), and a list of the indices of a high vowel in the input	
def height(input,FFC, highs):
	#Find the first index of a high vowel. 
	first=highs[0]
	#If only one index is in high, return the faithful candidate and the candidate with lowering.
	if len(highs)==1:
		if input[first]=="i" or input[first]=="I":
			low="e"
		else:
			low=input[first]
		return [(input,FFC), ("%s%s%s"%(input[:first],low,input[first+1:]),FFC)]
	#If there are multiple indices in highs, this function runs recursively.
	else:
		if input[first]=="i" or input[first]=="I":
			low="e"
		else:
			low=input[first]
		return height(input,FFC,highs[1:])+ height("%s%s%s"%(input[:first],low,input[first+1:]),FFC,highs[1:])
#Given the prefix and suffix strings, as well as the outpair (the surface form and correspondence relation).	
def vios(stem,suf, outpair):
	input="%s%s"%(stem,suf)
	output=outpair[0]
	corresp=outpair[1]
	#change the correspondence pairs from the correspondence relation, and turn them into a dictionary.
	crelation={}
	for x in corresp:
		crelation[x[0]]=x[1]

	
	viols=[]
	#Id[hi] add a violation if a vowel's input correspondent does not match its output correspondent in height
	Dhigh=0
	for x in range(len(input)):
		if x in crelation:
			#Add a violation if input segment is high and its output correspondent is not.
			if input[x] in HIGH and output[crelation[x]] not in HIGH:
				Dhigh+=1
				
	for x in range(len(input)):
		if x in crelation:
			#Add a violation if input segment is not high and its output correspondent is high.
			if input[x] not in HIGH and output[crelation[x]] in HIGH:
				Dhigh+=1

	#append the number of violations of Id[hi] to the violation vector.
	if Dhigh>=0:
		viols.append(Dhigh)
	else:
		viols.append(0)
	

	
	#Max-V add a violation if a vowel is not linked to an output correspondent.
	MinV=0
	
	for x in range(len(input)):#add one for each vowel with no correspondence to any output segments.
		if x not in crelation:
			MinV+=1
	#append the number of violations of Max-V to the violation vector.
	if MinV>=0:
		viols.append(MinV)
	else:
		viols.append(0)
		
	#*e Assign a violation for each output segment with [e]. 
	e=0
	for x in range(len(output)):
		if output[x]=="e":
			#if x==len(output)-1:
			#	e+=1
			#elif output[x+1]!=":":
			e+=1
			
	#append the number of violations of *e to the violation vector.
	viols.append(e)

	#maxv/in Assign a violation if the initial vowel is not linked to an output correspondent.
	MinV=0
	firstvow=-1
	outsyl=filter(lambda x: x in VOWELS+[':'], output)
	for x in range(len(input)):
		if input[x] in VOWELS: #Find the first segment in the input that is a vowel, save this index as firstvow.
			if firstvow==-1:
				firstvow=x
				#If the first vowel in the input has no output correspondent, assign one violation mark.
				if x not in crelation:
					MinV+=1
	#append the number of violations of maxv/in to the violation vector.
	if MinV>=0:
		viols.append(MinV)
	else:
		viols.append(0)
	
	#IDATR Assign a violation if a vowel the ATR value of the input is not the same as its output correspondent.
	DATR=0
	
	for x in range(len(input)):
		if x in crelation:
		#Add a violation if input segment is -ATR and its output correspondent is not.
			if input[x]=="I" and output[crelation[x]] !="I":
				DATR+=1
				
	for x in range(len(input)):
		if x in crelation:
		#Add a violation if input segment is +ATR and its output correspondent is not.
			if input[x]!="I" and output[crelation[x]]=="I":
				DATR+=1

	#append the number of violations of IdATR to the violation vector.	
	if DATR>=0:
		viols.append(DATR)
	else:
		viols.append(0)
	
	#DepV Assign a violation if an output vowel has no input correspondent
	PlsV=0
	for x in range(len(output)): #Iterate across the output segments and count a violation for each segment without an input correspondent
		if x not in crelation.itervalues():
			PlsV+=1
	
	#append the number of violations of DepV to the violation vector.
	if PlsV>=0:
		viols.append(PlsV)
	else:
		viols.append(0)


	#*I Assign a violation for each output segment with [I]. 
	I=0
	for x in range(len(output)):
		if output[x]=="I":
			I+=1
	#append the number of violations of *e to the violation vector.
	viols.append(I)
	
	#Ctk Assign a violation for each nonphonotactically illicit CCC sequence in the output
	ctk=0
	if 'tk' in output:
		if output[output.index('tk')-1] not in VOWELS+[":","n"]:
			ctk+=1
	if 'c3t' in output:
		ctk+=1
	if 'n5t' in output:
		ctk+=1
	if '*wl' in output:
		ctk+=1
	if 'mntk' in output:
		ctk+=1
	#append the number of violations of *Ctk to the violation vector.
	viols.append(ctk)
	#Id[hi]/in Assign a violation if the initial vowel of the input's high value  is not the same as its output correspondent.
	Dhigh=0
	x=firstvow
	if x in crelation:
		if input[x] in HIGH and output[crelation[x]] not in HIGH:
			Dhigh+=1
		if input[x] not in HIGH and output[crelation[x]] in HIGH:
			Dhigh+=1

	#append the number of violations of Id[hi]/in to the violation vector.
	if Dhigh>=0:
		viols.append(Dhigh)
	else:
		viols.append(0)
	#IdATR/in Assign a violation if the initial vowel of the input's  ATR value  is not the same as its output correspondent.
	Dhigh=0
	x=firstvow
	if x in crelation:
		if input[x] =="I" and output[crelation[x]]!="I":
			Dhigh+=1
		if input[x]=="I" and output[crelation[x]]!="I":
			Dhigh+=1

	#append the number of violations of ID[ATR]/Hi to the violation vector.
	if Dhigh>=0:
		viols.append(Dhigh)
	else:
		viols.append(0)

	return viols	
#Create candidates that epenthesize [a] . Take in an input string, FFC (a fully faithful correspondence relation), and blanks: a list of the indices of consonants followed by [t] or [C].	
def epenv(input,FFC,blanks):
	first=blanks[-1]
	#create a new correspondence relation to shift output indices to make up for the epenthesized vowel.
	new1=[]
	for j in FFC:
		i=j[1]
		k=j[0]
		if i<first:
			new1.append((k,i))
		elif i>=first:
			new1.append((k,i+1))
	newFFC=tuple(new1)
	#If there is just one index in blanks, return the fully faithful candidate as well as the candidate that inserts a vowel at first

	if len(blanks)==1:
		return [(input,FFC), ("%sa%s"%(input[:first],input[first:]),newFFC)]
	#If there are multiple indices in blanks, this function runs recursively.
	else:
		return epenv(input,FFC,blanks[:-1])+ epenv("%sa%s"%(input[:first],input[first:]),newFFC,blanks[:-1])

#Create candidates that delete [a] . Take in an input string, FFC (a fully faithful correspondence relation), and vowels: a list of the indices of all the vowels in the input.	
def delv(input,FFC, vowels):
	#If there are no indices in vowels, return just the faithful candidate.
	if len(vowels)==0:
		return [(input,FFC)]
	
	
	else:
		first=vowels[-1]
		#Develop a new correspondence relation shifted for the deletion of vowels
		new1=[]
		for j in FFC:
			i=j[1]
			k=j[0]
			if i<first:
				new1.append((k,i))
			elif i!=first:
				new1.append((k,i-1))
		new=tuple(new1)
		#Create deletion candidates if the vowels are marked (here /e/ and /I/)
		if input[first]=='e' or input[first]=='I':
			#If there is just one index in vowels, return the fully faithful candidate as well as the candidate that deletes the vowel at first
			if len(vowels)==1:
				return [(input,FFC), ("%s%s"%(input[:first],input[first+1:]),new)]
			#If there are multiple indices in vowels, this function runs recursively.
			else:
				return delv(input,FFC,vowels[:-1])+ delv("%s%s"%(input[:first],input[first+1:]),new,vowels[:-1])
		#Create deletion candidates for vowels in VV sequences.
		#If one of the vowels is initial in the input.
		elif first==0:
			if  input[first+1] in VOWELS: 
				#If there is just one index in vowels, return the fully faithful candidate as well as the candidate that deletes the vowel at first
				if len(vowels)==1:
					return [(input,FFC), ("%s%s"%(input[:first],input[first+1:]),new)]
				else:#If there are multiple indices in vowels, this function runs recursively.
					return delv(input,FFC,vowels[:-1])+ delv('%s%s'%(input[:first],input[first+1:]),new,vowels[:-1])
			else:
				return delv(input,FFC,vowels[:-1])
		#If one of the vowels i final.
		elif first==len(input)-1:
			if input[first-1] in VOWELS: 
				if len(vowels)==1:#If there are multiple indices in vowels, this function runs recursively.
					return [(input,FFC), ("%s%s"%(input[:first],input[first+1:]),new)]
				else:#If there are multiple indices in vowels, this function runs recursively.
					return delv(input,FFC,vowels[:-1])+ delv("%s%s"%(input[:first],input[first+1:]),new,vowels[:-1])
			else:
				return delv(input,FFC,vowels[:-1])
		#If the vowels are somewhere in he middle.
		elif input[first-1]  in VOWELS or input[first+1] in VOWELS: 
			if len(vowels)==1:
				return [(input,FFC), ('%s%s'%(input[:first],input[first+1:]),new)]
			else:
				return delv(input,FFC,vowels[:-1])+ delv('%s%s'%(input[:first],input[first+1:]),new,vowels[:-1])
			
		else:
			return [(input,FFC)]
			
#Take in an input string and a correspondence relation FFC. 		
def cands2(input,FFC):
	candset=[]
	#Create lowering of vowels
	highs=[]
	#Find the indices of each high vowel.
	for x in range(len(input)):
		if input[x] in HIGH:
			highs.append(x)
	if len(highs)==0:
		candset.append((input,FFC))
	else:# If there are high vowels call the height function which creates all permutations of lowering of high vowels.
		candset=candset+height(input,FFC,highs)
	#print candset
	#Create violations of Maxv 
	vs=[]
	#Find the indices of each vowel in the input word.
	for x in range(len(input)):
		if input[x] in VOWELS:
			vs.append(x)
	#Find the set of all permutations of markedness driven deletion 
	candset=candset+delv(input,FFC,vs)

	#Create rising of vowels
	lows=[]
	#Find indices of e and I
	for x in range(len(input)):
		if input[x] in ["e","I"]:
			lows.append(x)
	if len(lows)==0:
		candset.append((input,FFC))
	else: # If there are any vowels of this sort, return permutations of rising across the e and I vowels.
		candset=candset+rise(input,FFC,lows)
	# Create violations of DepV
	blanks=[]
	#Gather indices of consonants before t and C: blanks
	for x in range(len(input)):
		if x!=0:
			if input[x-1] not in VOWELS and input[x]=='t':
				blanks.append(x)
			if input[x-1] not in VOWELS and input[x]=='C':
				blanks.append(x)			
	if len(blanks)!=0: # Find all permutations of markedness driven epenthesis.
		candset=candset+epenv(input,FFC,blanks)
	
	return list(set(candset))	

canddict={}
#Given an input pair (stem, suffix), return a set of potential candidates. 
def cands(inpair):
	input="%s%s"%(inpair[0],inpair[1])
	sets=[]
	#Check if the set of candidates has been found already and is saved in canddict, so we don't have to run this too many times.
	if input in canddict:
		return canddict[input]
	else:
		#Create a correspondence relation with input and output.
		FFC=()
		for x in range(len(input)):
			FFC=FFC+((x,x),)
		#Call the cands2 function to generate a variety of output candidates.
		candset=cands2(input,FFC)
		# For each candidate in candset, add the set of candidates generated by cands2 for this candidate.
		for x in candset:
			sets=list(set(sets+cands2(x[0],x[1])))
		#Save the set of candidates in the canddict.
		canddict[input]=sets
		return sets
		
#Given inputpair and the WEIGHTS of constraints and find the probability of each candidate given the input prefix and suffix URs.
def probs(inputpair,WEIGHTS):
	candset=cands(inputpair)
	htotal=0
	harmdict={}
	#For each candidate find the harmony score and add up the total harmony score.
	for x in list(set(candset)):
		harmdict[x]=exp(harmony(inputpair,x,WEIGHTS))
		htotal+=harmdict[x]
	for x in harmdict: #Calculate the probability for each candidate
		if htotal<1e-150: #If all forms have very low probability give all 0 probability
			harmdict[x]=0
		else:
			harmdict[x]=harmdict[x]/htotal 
	return harmdict
	
#Calculate the L2 gaussian prior, the sum of the difference of each weight from it's bias point (mbias or 0) squared divided by the associated SIGMA value
def prior(WEIGHTS):
	total=0
	for i in range(len(WEIGHTS)):
		w=WEIGHTS[i]
		if CONTYPE[i]=='f':
			total+=w**2/(2*SIGMAF**2)
		else:
			total+=(w-mbias)**2/(2*SIGMAM**2)
	return total	
	
#Calculate the sum squared of all frequencies of underlying representations.	
def priorl(lexprob):
	sum=0
	for x in lexprob:
		sum+=x**2
	return sum
	
#Calculate the objective function for the given grammar.
def objective(parameters):
	#Separate the constraint weights (w) from the UR frequencies (lvect)
	w=parameters[:numcon]
	paramkey=paramlex[1]
	lvect=parameters[numcon:]
	URprobs={}
	#Build the dictionary of the frequency of each UR
	URfreqs={}
	for i in range(len(paramkey)):
		lam,morph=paramkey[i]
		val=lvect[i]
		if lam in URfreqs:
			URfreqs[lam][morph]=val
		else:
			URfreqs[lam]={morph:val}
	#Use pLEX to find the probability of each UR		
	URprobs=pLEX(URfreqs)

	probsum=0
	# For each piece of data find the stem and suffix morpheme tags
	for stemtag,suftag in inpout:
		likelihood=0
		surfaceform=inpout[(stemtag,suftag)]
		# for stem in stemtag and suf in suftag
		if stemtag not in URprobs:
			URprobs[stemtag]={LEX[stemtag].keys()[0]:1}
		for stem in URprobs[stemtag]:
			if suftag not in URprobs:
				URprobs[suftag]={LEX[suftag].keys()[0]:1}
			for suf in URprobs[suftag]:
				probps={}
				#pro is a dictionary of probability of each candidate given the input stem, suf
				pro=probs((stem,suf),w)
				for z in pro:
					if z[0] in probps:
						probps[z[0]]+=pro[z]
					else:
						probps[z[0]]=pro[z]
				#if the surfaceform is in probps add p(stem)p(suf)p(surfaceform|stem,suf) to likelihood	
				if surfaceform in probps:
					py=probps[surfaceform]
					likelihood+=URprobs[stemtag][stem]*URprobs[suftag][suf]*py
		#IF likelihood of the winner is super low, print error and mess up the objective
		if likelihood<=1e-150:
			if stemtag in URprobs:
				print LEX[stemtag]
				print "%s has gotten tooo small"%y
			probsum+=float('inf')
		else:	
			#add the negative log likelihood to probsum.
			probsum+=-log(likelihood)
	#As a failsafe, if some parameter is outside of its bounds, make the objective really big to let the algorithm know it messed up
	for x in parameters:
		if x<negative:
			return float("inf")
	#Return the sum of negative log likelihood with the prior on constraints and morpheme frequencies.
	return probsum+prior(w)+(sigma*priorl(lvect))

#Calculate the gradient of the objective function for each dimension
def gradient(parameters):
	grad=[]
	#Separate the constraint weights (w) from the UR frequencies (lvect)
	w=parameters[:numcon]
	paramkey=paramlex[1]
	lvect=parameters[numcon:]
	#Build the dictionary of the frequency of each UR
	URfreqs={}
	for i in range(len(paramkey)):
		lam,morph=paramkey[i]
		prob=lvect[i]
		if lam in URfreqs:
			URfreqs[lam][morph]=prob
		else:
			URfreqs[lam]={morph:prob}
	#Make a dictionary of probabilities of each UR
	URprobs=pLEX(URfreqs)
	
	canddict={}
	#START BY FINDING GRADIENTS FOR EACH CONSTRAINT WEIGHT
	for i in range(len(w)):
		sum=0
		likydict={}
		#For each piece of data, find the gradient for this constraint.
		for stemtag,suftag in inpout:
			surfaceform=inpout[(stemtag,suftag)]
			lik=0
			ts=0
			if stemtag not in URprobs:
				URprobs[stemtag]={LEX[stemtag].keys()[0]:1}
			if suftag not in URprobs:
				URprobs[suftag]={LEX[suftag].keys()[0]:1}
			#FOR EACH POSSIBLE Stem SUF PAIR
			for stem in URprobs[stemtag]:
				for suf in URprobs[suftag]:
					input="%s%s"%(stem,suf)
					#retrieve candidates for this output from canddict if possible. OR else create them. 
					if stem in canddict:
						if suf in canddict[stem]:
							candset=canddict[stem][suf]
						else:
							tmp=cands((stem,suf))
							canddict[stem][suf]=tmp
							candset=tmp
					else:
						tmp=cands((stem,suf))
						canddict[stem]={suf:tmp}
						candset=tmp
						
					# MAKE CANDY A LIST OF CANDIDATES WITHOUT correspondence relations.
					candy=[]
					for z in candset:
						if z[0] not in candy:
							candy.append(z[0])
					top=0
					bottom=0
					#IF surfaceform, the winning candidate, is in the list of candidates
					if surfaceform in candy:
						yharm=0
						vioy=0
						yprime=0
						#FOR ALL Z top=Sum(vios[z]eH(z))  bottom=SUM(eH(z)) yharm=Sum(eH(surfaceform)) yprime=Sum(vios[surfaceform]eH(surfaceform))  vioy=Sum(vios(surfaceform))
						for z in candset:
							zharm=exp(harmony((stem,suf),z,w))
							top+=vios(stem,suf,z)[i]*zharm
							bottom+=zharm
							if z[0]==surfaceform:
								yharm+=zharm
								yprime+=zharm*vios(stem,suf,z)[i]
								vioy+=vios(stem,suf,z)[i]
						#lk= p(stem)p(suf) Sum(eH(surfaceform))/Sum(eH(z))===likelihood(stem+suf)
						lk=URprobs[stemtag][stem]*URprobs[suftag][suf]*(yharm/bottom)
						lk2=(URprobs[stemtag][stem]*URprobs[suftag][suf]/bottom)
						#lik=Sum(likelihood(stem+suf))
						lik+=lk
						#ts=[sum(vios(z)eH(z))*Sum(eH(surfaceform))/Sum(eH(z))-Sum(vios[surfaceform]eH(surfaceform))]*likelihood(stem+suf)
						ts+=(top*yharm/bottom-yprime)*lk2
		
		
			if lik<1e-150:
				print "hmmm %s %s" %(surfaceform,lik)
				sum-=float('inf')
			else: #Add the gradient of the likelihood for this output form to the others.
				sum-=ts/lik
						
		#Find the gradient of the L2 Prior and add it to the sum of the gradient of the objective			
		wi=w[i]
		if CONTYPE[i]=='f':
			sum+=wi/(SIGMAF**2)
		else:
			sum+=(wi-mbias)/(SIGMAM**2)
		if wi<0:
			sum=-20
		grad.append(sum)

	#Find the gradient for each UR frequency.
	inputprobs={}
	for i in range(len(paramkey)):
		tag,morph= paramkey[i]
		#find total tag and morph val and dmorph
		totaltag=0
		for m in URfreqs[tag]:
			if m==morph:
				morphval=URfreqs[tag][m]
			totaltag+=URfreqs[tag][m]
	
		dmorph=float(totaltag-morphval)/(totaltag**2)	
		liky=0
		lastdict={}
		
		for tags in inpout:
			if tag in tags:
				stemtag,suftag=tags
				surfaceform=inpout[tags]
				top=0
				top2=0
				bottom=0
				#IF THIS MORPHEME IS A SUFFIX
				if tag in "0 1".split():
					# For each stem paired with this suffix, find the likelihood of the output form given this suffix.
					for stem in URprobs[stemtag]:
						pinput=(stem,morph)
						if pinput in inputprobs:
							outprob=inputprobs[pinput]
						else:
							outprob={}
							outtemp=probs((stem,morph),w)
							for z in outtemp:
								if z[0] in outprob:
									outprob[z[0]]+=outtemp[z]
								else:
									outprob[z[0]]=outtemp[z]
							inputprobs[pinput]=outprob
						#OUTPROB = sum(eH(surfaceform))/sum(eH(z)) for all surfaceform
						if surfaceform in outprob:
							pout=outprob[surfaceform]
						else:
							pout=0
						top+=(URprobs[stemtag][stem]*pout)
						for suf in URprobs[suftag]:
							input=(stem,suf)
							if input in inputprobs:
								outsuf=inputprobs[input]
							else:
								outsuf={}
								outtemp=probs((stem,suf),w)
								for z in outtemp:
									if z[0] in outsuf:
										outsuf[z[0]]+=outtemp[z]
									else:
										outsuf[z[0]]=outtemp[z]
								inputprobs[input]=outsuf
							if surfaceform in outsuf:
								bottom+=URprobs[stemtag][stem]*URprobs[suftag][suf]*outsuf[surfaceform]
								if suf!=morph:
									top2+=URprobs[stemtag][stem]*outsuf[surfaceform]*URfreqs[suftag][suf]/totaltag
					
				
				# IF THIS MORPHEME IS A STEM
				else:
					for suf in URprobs[suftag]:
						sinput=(morph,suf)
						if sinput in inputprobs:
							outprob=inputprobs[sinput]
						else:
							outprob={}
							outtemp=probs((morph,suf),w)
							for z in outtemp:
								if z[0] in outprob:
									outprob[z[0]]+=outtemp[z]
								else:
									outprob[z[0]]=outtemp[z]
							inputprobs[sinput]=outprob
					
						if surfaceform in outprob:
							pout=outprob[surfaceform]
						else:
							pout=0
						top+=URprobs[suftag][suf]*(pout)
						for stem in URprobs[stemtag]:
							input=(stem,suf)
							if input in inputprobs:
								outsuf=inputprobs[input]
							else:
								outsuf={}
								outtemp=probs((stem,suf),w)
								for z in outtemp:
									if z[0] in outsuf:
										outsuf[z[0]]+=outtemp[z]
									else:
										outsuf[z[0]]=outtemp[z]
								inputprobs[input]=outsuf
							if surfaceform in outsuf:
								bottom+=URprobs[stemtag][stem]*URprobs[suftag][suf]*outsuf[surfaceform]
								if stem!=morph:
									top2+=URprobs[suftag][suf]*outsuf[surfaceform]*URfreqs[stemtag][stem]/totaltag
				
								
				if bottom>1e-50:
					liky+=(dmorph*top-top2)/bottom
					
		gradi=-liky+(2*morphval)*sigma
		grad.append(gradi)
	printprob(parameters)
	
	print "---CONSTRAINTS------"
	printconstraints(parameters[:numcon])
	print "-----LEXICON-----"
	printdict(URprobs)
	
	
	print " gradient	%s"%np.linalg.norm(grad)
	print " objective %s"%objective(parameters)
	return np.asarray(grad)

#Given a dictionary d print those forms that have received less than .9 probability.
def printliky(d):
	y=0
	string=""
	if any(v < .9 for v in d.itervalues()):
		print "---SURFACE FORMS WITH LESS THAN .9 PROBABILITY----"	
	for x in d:
		if d[x]<.9:
			if len(x)<8:
				string+="	%s		%.2f,"%(x, d[x])
			else:
				string+="	%s	%.2f,"%(x, d[x])
			if y%3==2:
				print "%s"%(string)
				string=""
			y+=1
	
	print "%s" %string
	
	
#Print the probability of each output form given the grammar.	
def printprob(parameter):
	ew1=parameter[:numcon]
	lprevect=paramToVect(parameter[numcon:],lexvectkey,paramlex[1])
	URfreqs={}
	for i in range(len(lexvectkey)):
		lam,morph=lexvectkey[i]
		val=lprevect[i]
		if lam in URfreqs:
			URfreqs[lam][morph]=val
		else:
			URfreqs[lam]={morph:val}
	lexyprobs=pLEX(URfreqs)
	likylist={}
	for y in inpout:
		prob={}
		#print y
		lpre=y[0]
		lsuf=y[1]
		#print LEX
		for pre in LEX[lpre]:
			for suf in LEX[lsuf]:
				pbr=lexyprobs[lpre][pre]*lexyprobs[lsuf][suf]
				input="%s%s"%(pre,suf)
				proby=probs((pre,suf),ew1)
				for yum in proby:
					x=yum[0]
					if x in prob:
						prob[x]+=pbr*proby[yum]
					else:
						prob[x]=pbr*proby[yum]
		likylist[inpout[y]]=prob[inpout[y]]
	printliky(likylist)
#Print the weight of each constraint, with its name.
def printconstraints(d):
	string=""
	for x in range(numcon): 
		if len(CONSTRAINTS[x])<8:
			string+="	%s		%.2f,"%(CONSTRAINTS[x], d[x])
		else:
			string+="	%s	%.2f,"%(CONSTRAINTS[x], d[x])
		if x%4==3:
			print "%s"%(string)
			string=""
	print "%s"%(string)
#Given a dictionary of probabilities of URs for each morpheme, print them.
def printdict(d):
	for x in d:
		if len(d[x])>1:
			string=""
			for y in d[x]:
				if len(y)<8:
					string+="	%s		%.2f,"%(y, d[x][y])
				else:
					string+="	%s	%.2f,"%(y, d[x][y])
			print "%s"%(string)


def vectToParam(vect,key):
	param=[]
	paramkey=[]
	sum=0
	URfreqs={}
	for i in range(len(lexvectkey)):
		lam,morph=lexvectkey[i]
		val=vect[i]
		if lam in URfreqs:
			URfreqs[lam][morph]=val
		else:
			URfreqs[lam]={morph:val}
	for x in URfreqs:
		if len(URfreqs[x])>1:
			for y in URfreqs[x]:
				param.append(URfreqs[x][y])
				paramkey.append((x,y))

	return[param,paramkey]
#Given a dictionary of UR frequencies, return a vector.
def paramToVect(param,vkey,pkey):
	vect=[]
	
	for y in vkey:
		x=(y[0],y[1])
		if x in pkey:
			value=param[pkey.index(x)]
			vect.append(value)
			
		else:
			vect.append(1.0)
			
	return vect

paramlex=vectToParam(lexvect,lexvectkey)
test=wToParam(WEIGHTS,paramlex[0],paramlex[1])

paramlex=vectToParam(lexvect,lexvectkey)

#Create bounds for the optimize algorithm (don't let constraints or frequencies get negative or weights get above 200)
wbound=(0,200)
bounds=[wbound,]*(numcon)
probbond=(0,None)
lexlen=len(test[0])-len(WEIGHTS)
lexbond=[probbond,]*lexlen
bigbound=bounds+lexbond



#Run the SLSQP optimize algorithm on the objective function using the gradient function as a jacobian.
res=scipy.optimize.minimize(objective,np.asarray(test[0]), method='SLSQP',bounds=bigbound,jac=gradient,  options={'maxiter':1000,'ftol':1e-50, 'eps':1e-5, 'disp': True})

w1=res.x
print w1

ew1=[]
for x in w1[:numcon]:
	ew1.append(exp(x))
ew1=w1[:numcon]

lprevect=paramToVect(w1[numcon:],lexvectkey,paramlex[1])
print lprevect
lexvectout=[]
for x in lprevect:
	lexvectout.append(x)
	
URfreqs={}
for i in range(len(lexvectkey)):
	lam,morph=lexvectkey[i]
	val=lprevect[i]
	if lam in URfreqs:
		URfreqs[lam][morph]=val
	else:
		URfreqs[lam]={morph:val}
URprobs=pLEX(URfreqs)
lprobout=[]
for i in range(len(lexvectkey)):
	lam,morph=lexvectkey[i]
	lprobout.append(URprobs[lam][morph])

finallex=lprobout

#Print the results of the simulation.
print "CONSTRAINT WEIGHTINGS:"
for x in range(len(ew1)):
	print "%s		%s"%(CONSTRAINTS[x],ew1[x])
print "LEXICON PROBABILITIES:"
for x in range(len(finallex)):
	print "%s		%s"%(lexvectkey[x],finallex[x])
#Organize and print the URs for each morpheme, sorted by probability.
lexyprobs={}
for i in range(len(lexvectkey)):
	x=lexvectkey[i][0]
	y=lexvectkey[i][1]
	if x in lexyprobs:
		lexyprobs[x][y]=finallex[i]
	else:
		lexyprobs[x]={y:finallex[i]}

sortlexy={}
for x in lexyprobs:
	sortlexy[x]=[(k,lexyprobs[x][k]) for v,k in sorted([(v,k) for k,v in lexyprobs[x].items()],reverse=True)]
	print sortlexy[x][0]
	
#Calculate the Probability of each output form given its morpheme tags.
prod=1	
for y in inpout:
	prob={}
	#print y
	lpre=y[0]
	lsuf=y[1]
	#print LEX
	for pre in LEX[lpre]:
		for suf in LEX[lsuf]:
			pbr=lexyprobs[lpre][pre]*lexyprobs[lsuf][suf]
			input="%s%s"%(pre,suf)
			proby=probs((pre,suf),ew1)
			for yum in proby:
				x=yum[0]
				if x in prob:
					prob[x]+=pbr*proby[yum]
				else:
					prob[x]=pbr*proby[yum]
	sorted_prob = [(k,prob[k]) for v,k in sorted([(v,k) for k,v in prob.items()],reverse=True)]

	prod*=prob[inpout[y]]
#Print a warning if the output form is not produced with .9 probability, or is not the preferred output.
	if prob[inpout[y]]<.9:
		for pre in LEX[lpre]:
			for suf in LEX[lsuf]:
				input="%s%s"%(pre,suf)
				probinp=probs((pre,suf),ew1)
				sorts = [(k,probinp[k]) for v,k in sorted([(v,k) for k,v in probinp.items()],reverse=True)]
				#print "Aiming for %s"%inpout[y]
		print "%s:      %s"%(inpout[y], prob[inpout[y]])	
		if sorted_prob[0][1]!=prob[inpout[y]]:
			print "%s does not win:  %s"%(inpout[y], sorted_prob[:3])

#Print the overall probability of the data.
print "Probability of Data is %s"%prod
		